@@ -46,6 +46,22 @@
 extern int                              globus_i_gsi_gssapi_debug_level;
 extern FILE *                           globus_i_gsi_gssapi_debug_fstream;
 
+static
+int
+globus_l_gsi_gss_servername_callback(
+    SSL                                *SSL,
+    int                                *ad,
+    void                               *callback_arg);
+
+static
+int
+globus_l_gsi_gss_alpn_select_callback(
+    SSL                                *ssl,
+    const unsigned char               **out,
+    unsigned char                      *outlen,
+    const unsigned char                *in,
+    unsigned int                        inlen,
+    void                               *arg);
 /**
  * @defgroup globus_i_gsi_gss_utils Globus GSSAPI Internals
  *
@@ -229,6 +245,7 @@ globus_i_gsi_gss_create_and_fill_context(
     OM_uint32 *                         minor_status,
     gss_ctx_id_desc **                  context_handle_P,
     gss_OID                             mech,
+    const gss_name_t                    target_name,
     gss_cred_id_desc *                  cred_handle,
     const gss_cred_usage_t              cred_usage,
     OM_uint32                           req_flags)
@@ -239,6 +256,7 @@ globus_i_gsi_gss_create_and_fill_context(
     int                                 cb_index = -1;
     OM_uint32                           local_minor_status;
     char *                              certdir = NULL;
+    globus_bool_t                       allocated_context = GLOBUS_FALSE;
 
     GLOBUS_I_GSI_GSSAPI_DEBUG_ENTER;
 
@@ -259,6 +277,7 @@ globus_i_gsi_gss_create_and_fill_context(
         memset(context, 0, sizeof(gss_ctx_id_desc));
         *context_handle_P = context;
         context->ctx_flags = 0;
+        allocated_context = GLOBUS_TRUE;
     }
     else
     {
@@ -362,6 +381,35 @@ globus_i_gsi_gss_create_and_fill_context(
                 (gss_cred_id_t *) &context->cred_handle, 
                 NULL, 
                 NULL);
+
+            /* OK if that failed, if we are using SNI */
+            if (major_status != GSS_S_COMPLETE
+                && cred_usage != GSS_C_INITIATE
+                && context->sni_credentials == NULL
+                && getenv("X509_VHOST_CRED_DIR") != NULL)
+            {
+                globus_gsi_cred_handle_t    cred_handle = NULL;
+
+                local_result = globus_gsi_cred_handle_init(&cred_handle, NULL);
+                if (local_result != GLOBUS_SUCCESS)
+                {
+                    local_minor_status = local_result;
+                    major_status = GSS_S_FAILURE;
+                }
+                else
+                {
+                    major_status = globus_i_gsi_gss_create_cred(
+                        &local_minor_status,
+                        cred_usage,
+                        &context->cred_handle,
+                        &cred_handle,
+                        GLOBUS_TRUE);
+                    if (major_status != GSS_S_COMPLETE && cred_handle != NULL)
+                    {
+                        globus_gsi_cred_handle_destroy(cred_handle);
+                    }
+                }
+            }
         }
         
         if (GSS_ERROR(major_status))
@@ -379,6 +427,47 @@ globus_i_gsi_gss_create_and_fill_context(
         context->cred_obtained = 0;
         context->cred_handle = cred_handle;
     }
+    /* Set up the SNI callback if this is the accept side, and the app
+     * has provided an array of credentials to use
+     */
+    if (cred_usage != GSS_C_INITIATE)
+    {
+        if (context->sni_credentials == NULL
+            && getenv("X509_VHOST_CRED_DIR") != NULL)
+        {
+            size_t                      sni_creds_len = 0;
+
+            major_status = globus_i_gss_read_vhost_cred_dir(
+                minor_status,
+                NULL,
+                &context->sni_credentials,
+                &context->sni_credentials_count);
+            if(major_status != GLOBUS_SUCCESS)
+            {
+                goto free_callback_data;
+            }
+        }
+        if (context->sni_credentials_count > 0)
+        {
+            SSL_CTX_set_tlsext_servername_callback(
+                context->cred_handle->ssl_context,
+                globus_l_gsi_gss_servername_callback);
+
+            SSL_CTX_set_tlsext_servername_arg(
+                context->cred_handle->ssl_context,
+                context);
+        }
+    }
+#if OPENSSL_VERSION_NUMBER >= 0x10002000L
+    if (cred_usage != GSS_C_INITIATE
+        && context->alpn != NULL)
+    {
+        SSL_CTX_set_alpn_select_cb(
+            context->cred_handle->ssl_context,
+            globus_l_gsi_gss_alpn_select_callback,
+            context);
+    }
+#endif
 
     /* set the cert_dir in the callback data */
     local_result = 
@@ -417,11 +506,8 @@ globus_i_gsi_gss_create_and_fill_context(
         goto free_cert_dir;
     }
 
-    if(certdir)
-    {
-        free(certdir);
-        certdir = NULL;
-    }
+    free(certdir);
+    certdir = NULL;
     
     if(req_flags & GSS_C_GLOBUS_ALLOW_MISSING_SIGNING_POLICY)
     {
@@ -470,6 +556,24 @@ globus_i_gsi_gss_create_and_fill_context(
         major_status = GSS_S_FAILURE;
         goto free_cert_dir;
     }
+    /* Set the SNI name if we're the client and have a target */
+    if (cred_usage == GSS_C_INITIATE &&
+        target_name != GSS_C_NO_NAME &&
+        target_name->host_name != NULL)
+    {
+        SSL_set_tlsext_host_name(
+                    context->gss_ssl,
+                    target_name->host_name);
+    }
+#if OPENSSL_VERSION_NUMBER >= 0x10002000L
+    /* Set the ALPN if we're the client and have one set */
+    if (cred_usage == GSS_C_INITIATE
+        && context->alpn != NULL)
+    {
+        SSL_set_alpn_protos(
+            context->gss_ssl, context->alpn, context->alpn_length);
+    }
+#endif
 
     /* No longer setting SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS since it seemed
      * like a stop-gap measure to interoperate with broken SSL */
@@ -628,6 +732,7 @@ globus_i_gsi_gss_create_and_fill_context(
     if(context->gss_wbio)
     {
         BIO_free(context->gss_wbio);
+        context->gss_wbio = NULL;
     }
 
  free_rbio:
@@ -635,33 +740,34 @@ globus_i_gsi_gss_create_and_fill_context(
     if(context->gss_rbio)
     {
         BIO_free(context->gss_rbio);
+        context->gss_rbio = NULL;
     }
 
  free_cert_dir:
     
-    if(certdir)
-    {
-        globus_libc_free(certdir);
-    }
+    free(certdir);
+    certdir = NULL;
 
     if(context->cred_handle && context->cred_obtained)
     {
         gss_release_cred(&local_minor_status, 
                          (gss_cred_id_t *) &context->cred_handle);
+        context->cred_handle = NULL;
     }
 
  free_callback_data:
-
     if(context->callback_data)
     {
         globus_gsi_callback_data_destroy(context->callback_data);
+        context->callback_data = NULL;
     }
 
  free_proxy_handle:
     
-    if(context->proxy_handle)
+    if(context->proxy_handle != NULL)
     {
         globus_gsi_proxy_handle_destroy(context->proxy_handle);
+        context->proxy_handle = NULL;
     }
 
  free_peer_cred_handle:
@@ -669,20 +775,19 @@ globus_i_gsi_gss_create_and_fill_context(
     if(context->peer_cred_handle->cred_handle)
     {
         globus_gsi_cred_handle_destroy(context->peer_cred_handle->cred_handle);
+        context->peer_cred_handle->cred_handle = NULL;
     }
 
  free_peer_cred:
 
-    if(context->peer_cred_handle)
-    {
-        globus_libc_free(context->peer_cred_handle);
-    }
+    free(context->peer_cred_handle);
+    context->peer_cred_handle = NULL;
 
  free_context:
 
-    if(context)
+    if(allocated_context)
     {
-        globus_libc_free(context);
+        free(context);
         *context_handle_P = NULL;
     }
 
@@ -1363,7 +1468,8 @@ globus_i_gsi_gss_create_anonymous_cred(
     major_status = globus_i_gsi_gssapi_init_ssl_context(
         &local_minor_status,
         (gss_cred_id_t) newcred,
-        GLOBUS_I_GSI_GSS_ANON_CONTEXT);
+        GLOBUS_I_GSI_GSS_ANON_CONTEXT,
+        GLOBUS_FALSE);
     if(GSS_ERROR(major_status))
     {
         GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
@@ -1437,7 +1543,8 @@ globus_i_gsi_gss_cred_read_bio(
     major_status = globus_i_gsi_gss_create_cred(&local_minor_status,
                                                 cred_usage,
                                                 cred_id_handle, 
-                                                &local_cred_handle);
+                                                &local_cred_handle,
+                                                GLOBUS_FALSE);
     
     if(GSS_ERROR(major_status))
     {
@@ -1470,6 +1577,7 @@ globus_i_gsi_gss_cred_read(
     globus_gsi_cred_handle_t            local_cred_handle;
     OM_uint32                           local_minor_status;
     OM_uint32                           major_status = GSS_S_COMPLETE;
+    int                                 rc = 0;
 
     GLOBUS_I_GSI_GSSAPI_DEBUG_ENTER;
 
@@ -1484,8 +1592,23 @@ globus_i_gsi_gss_cred_read(
         goto exit;
     }
 
+
     local_result = globus_gsi_cred_read(local_cred_handle, 
                                         (X509_NAME *) desired_subject);
+#ifndef WIN32
+    if(local_result != GLOBUS_SUCCESS
+        && getuid() == 0
+        && globus_i_gsi_gssapi_vhost_cred_owner != 0)
+    {
+        rc = seteuid(globus_i_gsi_gssapi_vhost_cred_owner);
+        local_result = globus_gsi_cred_read(local_cred_handle, 
+                                            (X509_NAME *) desired_subject);
+        if (rc == 0)
+        {
+            seteuid(0);
+        }
+    }
+#endif
     if(local_result != GLOBUS_SUCCESS)
     {
         GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
@@ -1498,7 +1621,8 @@ globus_i_gsi_gss_cred_read(
     major_status = globus_i_gsi_gss_create_cred(&local_minor_status,
                                                 cred_usage,
                                                 cred_handle, 
-                                                &local_cred_handle);
+                                                &local_cred_handle,
+                                                GLOBUS_FALSE);
     if(GSS_ERROR(major_status))
     {
         GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
@@ -1538,7 +1662,8 @@ globus_i_gsi_gss_create_cred(
     OM_uint32 *                         minor_status,
     const gss_cred_usage_t              cred_usage,
     gss_cred_id_t *                     output_cred_handle_P,
-    globus_gsi_cred_handle_t *          cred_handle)
+    globus_gsi_cred_handle_t *          cred_handle,
+    globus_bool_t                       sni_context)
 {
     gss_cred_id_desc **                 output_cred_handle = 
         (gss_cred_id_desc **) output_cred_handle_P;
@@ -1584,7 +1709,8 @@ globus_i_gsi_gss_create_cred(
     major_status = globus_i_gsi_gssapi_init_ssl_context(
         &local_minor_status,
         (gss_cred_id_t) newcred,
-        GLOBUS_I_GSI_GSS_DEFAULT_CONTEXT);
+        GLOBUS_I_GSI_GSS_DEFAULT_CONTEXT,
+        sni_context);
     if(GSS_ERROR(major_status))
     {
         GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
@@ -1594,6 +1720,10 @@ globus_i_gsi_gss_create_cred(
         goto error_exit;
     }
 
+    if (sni_context)
+    {
+        goto skip_for_sni;
+    }
     local_result = globus_gsi_cred_get_cert_type(
         newcred->cred_handle, 
         &cert_type);
@@ -1668,6 +1798,7 @@ globus_i_gsi_gss_create_cred(
     }
             
     
+skip_for_sni:
     *output_cred_handle = newcred;
 
  error_exit:
@@ -2119,7 +2250,8 @@ OM_uint32
 globus_i_gsi_gssapi_init_ssl_context(
     OM_uint32 *                         minor_status,
     gss_cred_id_t                       credential,
-    globus_i_gsi_gss_context_type_t     anon_ctx)
+    globus_i_gsi_gss_context_type_t     anon_ctx,
+    globus_bool_t                       sni_context)
 {
     X509 *                              client_cert = NULL;
     EVP_PKEY *                          client_key = NULL;
@@ -2279,114 +2411,117 @@ globus_i_gsi_gssapi_init_ssl_context(
     
     if(anon_ctx != GLOBUS_I_GSI_GSS_ANON_CONTEXT)
     {
-        local_result = globus_gsi_cred_get_cert(cred_handle->cred_handle,
-                                                &client_cert);
-        if(local_result != GLOBUS_SUCCESS)
+        if (!sni_context)
         {
-            GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
-                minor_status, local_result,
-                GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL);
-            major_status = GSS_S_FAILURE;
-            goto exit;
-        }
-
-        if(!client_cert)
-        {
-            GLOBUS_GSI_GSSAPI_ERROR_RESULT(
-                minor_status,
-                GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL,
-                (_GGSL("The GSI credential's certificate has not been set.")));
-            major_status = GSS_S_FAILURE;
-            goto exit;
-        }
-
-        if(!SSL_CTX_use_certificate(cred_handle->ssl_context, 
-                                    client_cert))
-        {
-            GLOBUS_GSI_GSSAPI_OPENSSL_ERROR_RESULT(
-                minor_status,
-                GLOBUS_GSI_GSSAPI_ERROR_WITH_OPENSSL,
-                (_GGSL("Couldn't set the certificate to "
-                 "be used for the SSL context")));
-            major_status = GSS_S_FAILURE;
-            goto exit;
-        }
-
-        local_result = globus_gsi_cred_get_key(cred_handle->cred_handle,
-                                               &client_key);
-        if(local_result != GLOBUS_SUCCESS)
-        {
-            GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
-                minor_status, local_result,
-                GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL);
-            major_status = GSS_S_FAILURE;
-            goto exit;
-        }
-
-        if(!client_key)
-        {
-            GLOBUS_GSI_GSSAPI_ERROR_RESULT(
-                minor_status,
-                GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL,
-                (_GGSL("The GSI credential's private key has not been set.")));
-            major_status = GSS_S_FAILURE;
-            goto exit;
-        }
-
-        if(!SSL_CTX_use_PrivateKey(cred_handle->ssl_context, client_key))
-        {
-            GLOBUS_GSI_GSSAPI_OPENSSL_ERROR_RESULT(
-                minor_status,
-                GLOBUS_GSI_GSSAPI_ERROR_WITH_OPENSSL,
-                (_GGSL("Couldn't set the private key to "
-                 "be used for the SSL context")));
-            major_status = GSS_S_FAILURE;
-            goto exit;
-        }
-            
-        local_result = globus_gsi_cred_get_cert_chain(cred_handle->cred_handle,
-                                                      &client_cert_chain);
-        if(local_result != GLOBUS_SUCCESS)
-        {
-            GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
-                minor_status, local_result,
-                GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL);
-            major_status = GSS_S_FAILURE;
-            goto exit;
-        }
-
-        if(client_cert_chain)
-        {
-            int                         index;
-            X509 *                      tmp_cert = NULL;
-            for(index = 0; index < sk_X509_num(client_cert_chain); ++index)
+            local_result = globus_gsi_cred_get_cert(cred_handle->cred_handle,
+                                                    &client_cert);
+            if(local_result != GLOBUS_SUCCESS)
             {
-                tmp_cert = X509_dup(sk_X509_value(client_cert_chain, index));
-                if(!X509_STORE_add_cert(
-                       SSL_CTX_get_cert_store(cred_handle->ssl_context),
-                       tmp_cert))
+                GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
+                    minor_status, local_result,
+                    GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL);
+                major_status = GSS_S_FAILURE;
+                goto exit;
+            }
+
+            if(!client_cert)
+            {
+                GLOBUS_GSI_GSSAPI_ERROR_RESULT(
+                    minor_status,
+                    GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL,
+                    (_GGSL("The GSI credential's certificate has not been set.")));
+                major_status = GSS_S_FAILURE;
+                goto exit;
+            }
+
+            if(!SSL_CTX_use_certificate(cred_handle->ssl_context, 
+                                        client_cert))
+            {
+                GLOBUS_GSI_GSSAPI_OPENSSL_ERROR_RESULT(
+                    minor_status,
+                    GLOBUS_GSI_GSSAPI_ERROR_WITH_OPENSSL,
+                    (_GGSL("Couldn't set the certificate to "
+                     "be used for the SSL context")));
+                major_status = GSS_S_FAILURE;
+                goto exit;
+            }
+
+            local_result = globus_gsi_cred_get_key(cred_handle->cred_handle,
+                                                   &client_key);
+            if(local_result != GLOBUS_SUCCESS)
+            {
+                GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
+                    minor_status, local_result,
+                    GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL);
+                major_status = GSS_S_FAILURE;
+                goto exit;
+            }
+
+            if(!client_key)
+            {
+                GLOBUS_GSI_GSSAPI_ERROR_RESULT(
+                    minor_status,
+                    GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL,
+                    (_GGSL("The GSI credential's private key has not been set.")));
+                major_status = GSS_S_FAILURE;
+                goto exit;
+            }
+
+            if(!SSL_CTX_use_PrivateKey(cred_handle->ssl_context, client_key))
+            {
+                GLOBUS_GSI_GSSAPI_OPENSSL_ERROR_RESULT(
+                    minor_status,
+                    GLOBUS_GSI_GSSAPI_ERROR_WITH_OPENSSL,
+                    (_GGSL("Couldn't set the private key to "
+                     "be used for the SSL context")));
+                major_status = GSS_S_FAILURE;
+                goto exit;
+            }
+                
+            local_result = globus_gsi_cred_get_cert_chain(cred_handle->cred_handle,
+                                                          &client_cert_chain);
+            if(local_result != GLOBUS_SUCCESS)
+            {
+                GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
+                    minor_status, local_result,
+                    GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL);
+                major_status = GSS_S_FAILURE;
+                goto exit;
+            }
+
+            if(client_cert_chain)
+            {
+                int                         index;
+                X509 *                      tmp_cert = NULL;
+                for(index = 0; index < sk_X509_num(client_cert_chain); ++index)
                 {
+                    tmp_cert = X509_dup(sk_X509_value(client_cert_chain, index));
+                    if(!X509_STORE_add_cert(
+                           SSL_CTX_get_cert_store(cred_handle->ssl_context),
+                           tmp_cert))
+                    {
+                        /* need to free to reduce ref count */
+                        X509_free(tmp_cert);
+                        if ((ERR_GET_REASON(ERR_peek_error()) ==
+                             X509_R_CERT_ALREADY_IN_HASH_TABLE))
+                        {
+                            ERR_clear_error();
+                            break;
+                        }
+                        else
+                        {
+                            GLOBUS_GSI_GSSAPI_OPENSSL_ERROR_RESULT(
+                                minor_status,
+                                GLOBUS_GSI_GSSAPI_ERROR_WITH_OPENSSL,
+                                (_GGSL("Couldn't add certificate to the SSL context's "
+                                 "certificate store.")));
+                            major_status = GSS_S_FAILURE;
+                            goto exit;
+                        }
+                    }
                     /* need to free to reduce ref count */
                     X509_free(tmp_cert);
-                    if ((ERR_GET_REASON(ERR_peek_error()) ==
-                         X509_R_CERT_ALREADY_IN_HASH_TABLE))
-                    {
-                        ERR_clear_error();
-                        break;
-                    }
-                    else
-                    {
-                        GLOBUS_GSI_GSSAPI_OPENSSL_ERROR_RESULT(
-                            minor_status,
-                            GLOBUS_GSI_GSSAPI_ERROR_WITH_OPENSSL,
-                            (_GGSL("Couldn't add certificate to the SSL context's "
-                             "certificate store.")));
-                        major_status = GSS_S_FAILURE;
-                        goto exit;
-                    }
                 }
-                /* need to free to reduce ref count */
-                X509_free(tmp_cert);
             }
         }
     }
@@ -2684,5 +2819,277 @@ out:
     return major_status;
 }
 
+static
+int
+globus_l_gsi_gss_servername_callback(
+    SSL                                *s,
+    int                                *ad,
+    void                               *callback_arg)
+{
+    gss_ctx_id_t                        context = callback_arg;
+    gss_cred_id_t                       credential = NULL;
+    const char                         *servername = NULL;
+    gss_name_t                          imported_name = GSS_C_NO_NAME;
+    globus_result_t                     local_result = GLOBUS_SUCCESS;
+    OM_uint32                           major_status = GSS_S_COMPLETE;
+    OM_uint32                           local_minor_status = GLOBUS_SUCCESS;
+    X509                               *client_cert = NULL;
+    EVP_PKEY                           *client_key = NULL;
+    STACK_OF(X509)                     *client_cert_chain = NULL;
+    globus_gsi_cert_utils_cert_type_t   cert_type;
+    
+    servername = SSL_get_servername(s, TLSEXT_NAMETYPE_host_name);
+    if(GLOBUS_I_GSI_GSSAPI_DEBUG(3))
+    {
+        BIO *                       debug_bio;
+        fprintf(globus_i_gsi_gssapi_debug_fstream,
+                "SNI callback: servername = %s\n",
+                (servername != NULL) ? servername : "NULL");
+    }
+
+    if (servername == NULL)
+    {
+        if (globus_gsi_cred_get_cert_type(
+                context->cred_handle->cred_handle,
+                &cert_type) == GLOBUS_SUCCESS)
+        {
+            /* No SNI, use default credential */
+            return SSL_TLSEXT_ERR_OK;
+        }
+        else
+        {
+            credential = context->sni_credentials[0];
+            goto use_any;
+        }
+    }
+    context->sni_servername = strdup(servername);
+
+    major_status = gss_import_name(
+            &local_minor_status,
+            &(gss_buffer_desc)
+            {
+                .value = (void *) servername,
+                .length = strlen(servername),
+            },
+            GLOBUS_GSS_C_NT_HOST_IP,
+            &imported_name);
+
+    if (major_status != GSS_S_COMPLETE)
+    {
+        return SSL_TLSEXT_ERR_ALERT_FATAL;
+    }
+
+    for (size_t i = 0 ; i < context->sni_credentials_count; i++)
+    {
+        int                             name_equal = 0;
+
+        major_status = gss_compare_name(
+                &local_minor_status,
+                imported_name,
+                context->sni_credentials[i]->globusid,
+                &name_equal);
+
+        if (major_status == GSS_S_COMPLETE && name_equal)
+        {
+            credential = context->sni_credentials[i];
+            break;
+        }
+    }
+    gss_release_name(
+            &local_minor_status,
+            &imported_name);
+    imported_name = GSS_C_NO_NAME;
+
+    if (major_status != GSS_S_COMPLETE)
+    {
+        return SSL_TLSEXT_ERR_ALERT_FATAL;
+    }
+
+    if (credential == NULL)
+    {
+        return SSL_TLSEXT_ERR_NOACK;
+    }
+
+use_any:
+    /* Replace the credential used by the context with the newly acquired
+     * one
+     */
+    if (context->cred_obtained)
+    {
+        major_status = gss_release_cred(
+            &local_minor_status,
+            &context->cred_handle);
+
+        if (GSS_ERROR(major_status))
+        {
+            goto exit;
+        }
+    }
+
+    context->cred_handle = credential;
+    context->cred_obtained = GLOBUS_FALSE;
+    credential = NULL;
+
+    local_result = globus_gsi_cred_get_cert(
+            context->cred_handle->cred_handle, &client_cert);
+    if (local_result != GLOBUS_SUCCESS)
+    {
+        GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
+            &local_minor_status, local_result,
+            GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL);
+        major_status = GSS_S_FAILURE;
+        goto exit;
+    }
+
+    if (client_cert == NULL)
+    {
+        GLOBUS_GSI_GSSAPI_ERROR_RESULT(
+            &local_minor_status,
+            GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL,
+            (_GGSL("The GSI credential's certificate has not been set.")));
+        major_status = GSS_S_FAILURE;
+        goto exit;
+    }
+
+    if (!SSL_CTX_use_certificate(
+                context->cred_handle->ssl_context, 
+                client_cert))
+    {
+        GLOBUS_GSI_GSSAPI_OPENSSL_ERROR_RESULT(
+            &local_minor_status,
+            GLOBUS_GSI_GSSAPI_ERROR_WITH_OPENSSL,
+            (_GGSL("Couldn't set the certificate to "
+             "be used for the SSL context")));
+        major_status = GSS_S_FAILURE;
+        goto exit;
+    }
+
+    local_result = globus_gsi_cred_get_key(context->cred_handle->cred_handle,
+                                           &client_key);
+    if(local_result != GLOBUS_SUCCESS)
+    {
+        GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
+            &local_minor_status, local_result,
+            GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL);
+        major_status = GSS_S_FAILURE;
+        goto exit;
+    }
+
+    if (client_key == NULL)
+    {
+        GLOBUS_GSI_GSSAPI_ERROR_RESULT(
+            &local_minor_status,
+            GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL,
+            (_GGSL("The GSI credential's private key has not been set.")));
+        major_status = GSS_S_FAILURE;
+        goto exit;
+    }
+
+    if(!SSL_CTX_use_PrivateKey(context->cred_handle->ssl_context, client_key))
+    {
+        GLOBUS_GSI_GSSAPI_OPENSSL_ERROR_RESULT(
+            &local_minor_status,
+            GLOBUS_GSI_GSSAPI_ERROR_WITH_OPENSSL,
+            (_GGSL("Couldn't set the private key to "
+             "be used for the SSL context")));
+        major_status = GSS_S_FAILURE;
+        goto exit;
+    }
+        
+    local_result = globus_gsi_cred_get_cert_chain(
+            context->cred_handle->cred_handle,
+            &client_cert_chain);
+    if(local_result != GLOBUS_SUCCESS)
+    {
+        GLOBUS_GSI_GSSAPI_ERROR_CHAIN_RESULT(
+            &local_minor_status, local_result,
+            GLOBUS_GSI_GSSAPI_ERROR_WITH_GSI_CREDENTIAL);
+        major_status = GSS_S_FAILURE;
+        goto exit;
+    }
+
+    if(client_cert_chain)
+    {
+        int                         index;
+        X509 *                      tmp_cert = NULL;
+        for(index = 0; index < sk_X509_num(client_cert_chain); ++index)
+        {
+            tmp_cert = X509_dup(sk_X509_value(client_cert_chain, index));
+            if(!X509_STORE_add_cert(
+                   SSL_CTX_get_cert_store(context->cred_handle->ssl_context),
+                   tmp_cert))
+            {
+                /* need to free to reduce ref count */
+                X509_free(tmp_cert);
+                if ((ERR_GET_REASON(ERR_peek_error()) ==
+                     X509_R_CERT_ALREADY_IN_HASH_TABLE))
+                {
+                    ERR_clear_error();
+                    break;
+                }
+                else
+                {
+                    GLOBUS_GSI_GSSAPI_OPENSSL_ERROR_RESULT(
+                        &local_minor_status,
+                        GLOBUS_GSI_GSSAPI_ERROR_WITH_OPENSSL,
+                        (_GGSL("Couldn't add certificate to the SSL context's "
+                         "certificate store.")));
+                    major_status = GSS_S_FAILURE;
+                    goto exit;
+                }
+            }
+            /* need to free to reduce ref count */
+            X509_free(tmp_cert);
+        }
+    }
+    SSL_set_SSL_CTX(context->gss_ssl, context->cred_handle->ssl_context);
+exit:
+    X509_free(client_cert);
+    EVP_PKEY_free(client_key);
+    sk_X509_pop_free(client_cert_chain, X509_free);
+
+    if (major_status != GSS_S_COMPLETE)
+    {
+        return SSL_TLSEXT_ERR_ALERT_FATAL;
+    }
+    else
+    {
+        return SSL_TLSEXT_ERR_OK;
+    }
+}
+
+#if OPENSSL_VERSION_NUMBER >= 0x10002000L
+int
+globus_l_gsi_gss_alpn_select_callback(
+    SSL                                *ssl,
+    const unsigned char               **out,
+    unsigned char                      *outlen,
+    const unsigned char                *in,
+    unsigned int                        inlen,
+    void                               *arg)
+{
+    int                                 rc = 0;
+    gss_ctx_id_desc*                    context = arg;
+    unsigned char                      *tmpout = NULL;
+
+    rc = SSL_select_next_proto(
+        &tmpout,
+        outlen,
+        context->alpn,
+        context->alpn_length,
+        in,
+        inlen);
+
+    if (rc == OPENSSL_NPN_NEGOTIATED)
+    {
+        *out = tmpout;
+        return SSL_TLSEXT_ERR_OK;
+    }
+    else
+    {
+        return SSL_TLSEXT_ERR_NOACK;
+    }
+}
+#endif
 
 #endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */
